// Author: Vladimir Savic
// E-mail: vladimirsvc@yahoo.com

// StateMachineUnitTest.cpp : Defines the entry point for the console application.
//
#include "stdafx.h"

#define BUFFER_SIZE	         29
#define NUMBER_OF_ITERATIONS 50

// ***************************************************************************************************************************************************
// *  TRANSITION MATRIX STRUCTURE:                                                                                                                   *
// *    <-                        NUMBER_OF_TM_COLUMNS                              ->                                                               *
// *    <-            3 bytes          -><-       external commands 4 bytes         ->                                                               *
// *    ------------------------------------------------------------------------------                                                               *
// *    |    DT    |     PS   |    TPS   |   Cmd1   |   Cmd2   |   Cmd3   |   Cmd4   |                                                               *
// *    ------------------------------------------------------------------------------                                                               *
// * SA |<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|                                                               *
// *    ------------------------------------------------------------------------------                                                               *
// * SB |<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|                                                               *
// *    ------------------------------------------------------------------------------                                                               *
// * SC |<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|                                                               *
// *    ------------------------------------------------------------------------------                                                               *
// * SD |<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|                                                               *
// *    ------------------------------------------------------------------------------                                                               *
// *                                                                                                                                                 *
// *  TRANSITION MATRIX LEGEND:                                                                                                                      *
// *  PS   - Previous state                                                                                                                          *
// *  TPS  - Transition by PS  (state machine goes directly to next state only if PS condition is satisfied)                                         *
// *  DT   - Direct transition (when default method is executed, state machine goes directly to another state)                                       *
// *  CmdX - where X belongs to the set of N numbers, present external commands, also called signals/events                                          *
// *  Sx   - where x belongs to the set of N numbers, present state of the state machine                                                             *
// *  0xFF - (255 decimal) means "nothing" (like void in C language) - 0xFF cannot be used for state/cmd encoding                                    *
// *         for example: if DT is set to 0xFF, state does not has direct transition to another state                                                *
// *                      if PS is set to 0xFF, state machine does not look at TPS - transition to another state does not depends on PS              *
// *                                                                                                                                                 *
// *  TRANSITION RULES:                                                                                                                              *
// *  -----------------                                                                                                                              *
// *  Transition types:                                                                                                                              *
// *  PS, TPS  - conditional transition, transition to another state depends on previous state.                                                      *
// *             If previous state is defined TPS field determines next state, all external commands are excluded                                    *
// *  DT       - direct transition, if direct transition is set all external commands are excluded including conditional transition                  *
// *  CmdX     - external command transition. External transitions are included only in case when DT is set to 0xFF as well as PS                    *
// *  Transition priorities (smaller number means higher priority level):                                                                            *
// *  DT   - 1 level                                                                                                                                 *
// *  PS   - 2 level                                                                                                                                 *
// *  CmdX - 3 level                                                                                                                                 *
// *                                                                                                                                                 *
// *  TRANSITION MATRIX FILEDS DEFINITION:                                                                                                           *
// *  each transition matrix filed is determined by 8 bits (BYTE)                                                                                    *
// *                                                                                                                                                 *
// *  ENCODING:                                                                                                                                      *
// *  each state/cmd must be encoded (identified) with 8 bits                                                                                        *
// *                                                                                                                                                 *
// *  states encoding:                                                                                                                               *
// *  SA -> 0x00 (first  transition matrix row, index 0)                                                                                             *
// *  SB -> 0x01 (second transition matrix row, index 1)                                                                                             *
// *  SC -> 0x02 (third  transition matrix row, index 2)                                                                                             *
// *  SD -> 0x03 (fourth transition matrix row, index 3)                                                                                             *
// *                                                                                                                                                 *
// *  commands encoding (0x00, 0x01 and 0x02 cannot be used for command encoding - those codes are reserved for DT, PS and TPS respectively):        *
// *  encoding for external commands starts from 0x03 ...                                                                                            *
// *  ---------------------------------------------------------------------------------------------------------------------------------------        *
// *  DT   -> always has 0x00 code, transition matrix column index 0                                                                                 *
// *  PS   -> always has 0x01 code, transition matrix column index 1                                                                                 *
// *  TPS  -> always has 0x02 code, transition matrix column index 2                                                                                 *
// *  ---------------------------------------------------------------------------------------------------------------------------------------        *
// *  Cmd1 -> 0x03                                                                                                                                   *
// *  Cmd2 -> 0x04                                                                                                                                   *
// *  Cmd3 -> 0x05                                                                                                                                   *
// *  Cmd4 -> 0x06                                                                                                                                   *
// *                                                                                                                                                 *
// *  SERIAL COMMUNICATION BUFFER STREAM (for this example):                                                                                         *
// *  28 bytes buffer required for transition matrix - TM                                                                                            *
// *  1  byte required for start state identificator - S0                                                                                            *
// *  --------------------------                                                                                                                     *
// *  |    S0    |     TM      |                                                                                                                     *
// *  --------------------------                                                                                                                     *
// *  |<- BYTE ->|<- 28xBYTES ->                                                                                                                     *
// *                                                                                                                                                 *
// *  TM STREAM:                                                                                                                                     *
// *  ----------------------------------------------------------------------------------------------------------                                     *
// *  |    DT    |     PS   |    TPS   |   Cmd1   |   Cmd2   |   Cmd3   |   Cmd4   |           ...             |                                     *
// *  ----------------------------------------------------------------------------------------------------------                                     *
// *  |<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|           ...             |                                     *
// *  ----------------------------------------------------------------------------------------------------------                                     *
// *  |<-                                 SA State                               ->|<- other states: SB..SD  ->|                                     *
// *  ----------------------------------------------------------------------------------------------------------                                     *
// ***************************************************************************************************************************************************

// *******************************
// * Global variable declaration *
// *******************************

// First Injection System Behavior:
// --------------------------------
// S0 = SA
//		     <-                        NUMBER_OF_TM_COLUMNS                              ->
//		     <-            3 bytes          -><-       external commands 4 bytes         ->
//		     ------------------------------------------------------------------------------
//		     |    DT    |     PS   |    TPS   |   Cmd1   |   Cmd2   |   Cmd3   |   Cmd4   |
//		     ------------------------------------------------------------------------------
// SA	0x00 |   0xFF   |   0xFF   |   0xFF   |   0x02   |   0xFF   |   0xFF   |   0xFF   |
//		     ------------------------------------------------------------------------------
// SB	0x01 |   0x00   |   0xFF   |   0xFF   |   0xFF   |   0xFF   |   0xFF   |   0xFF   |
//		     ------------------------------------------------------------------------------
// SC	0x02 |   0xFF   |   0x00   |   0x01   |   0xFF   |   0xFF   |   0xFF   |   0x03   |
//		     ------------------------------------------------------------------------------
// SD	0x03 |   0xFF   |   0xFF   |   0xFF   |   0xFF   |   0xFF   |   0x02   |   0xFF   |
//		     ------------------------------------------------------------------------------
//               0x00       0x01       0x02       0x03       0x04       0x05       0x06

BYTE firstInjectionSystemBehavior [BUFFER_SIZE] = { 0x00,                                      // S0    Start state
	                                                0xFF, 0xFF, 0xFF, 0x02, 0xFF, 0xFF, 0xFF,  // SA	0x00
													0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  // SB	0x01
													0xFF, 0x00, 0x01, 0xFF, 0xFF, 0xFF, 0x03,  // SC	0x02
													0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x02, 0xFF   // SD	0x03
                                                  };
// Second Injection System Behavior:
// --------------------------------
// S0 = SA
//		     <-                        NUMBER_OF_TM_COLUMNS                              ->
//		     <-            3 bytes          -><-       external commands 4 bytes         ->
//		     ------------------------------------------------------------------------------
//		     |    DT    |     PS   |    TPS   |   Cmd1   |   Cmd2   |   Cmd3   |   Cmd4   |
//		     ------------------------------------------------------------------------------
// SA	0x00 |   0xFF   |   0x03   |   0x00   |   0xFF   |   0x01   |   0xFF   |   0x02   |
//		     ------------------------------------------------------------------------------
// SB	0x01 |   0x03   |   0xFF   |   0xFF   |   0xFF   |   0xFF   |   0xFF   |   0xFF   |
//		     ------------------------------------------------------------------------------
// SC	0x02 |   0xFF   |   0xFF   |   0xFF   |   0x00   |   0xFF   |   0xFF   |   0xFF   |
//		     ------------------------------------------------------------------------------
// SD	0x03 |   0xFF   |   0xFF   |   0xFF   |   0xFF   |   0xFF   |   0x00   |   0xFF   |
//		     ------------------------------------------------------------------------------
//               0x00       0x01       0x02       0x03       0x04       0x05       0x06

BYTE secondInjectionSystemBehavior [BUFFER_SIZE] = { 0x02,                                      // S0   Start state
	                                                 0xFF, 0x03, 0x00, 0xFF, 0x01, 0xFF, 0x02,  // SA	0x00
													 0x03, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  // SB	0x01
													 0xFF, 0xFF, 0xFF, 0x00, 0xFF, 0xFF, 0xFF,  // SC	0x02
													 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0xFF   // SD	0x03
                                                   };

BYTE* transitionMatrixBuffer;

BYTE counter                 ;
BYTE command                 ;
BYTE portValueState          ;
BYTE portValuePrevState      ;
BYTE portValueIntHandler     ;
BYTE portValuePrevIntHandler ;

TransitionMatrix     transitionMatrix;
StateMachineAbstract stateMachine    ;

// ****************************************
// * Preprocessor and function prototypes *
// ****************************************
#define PRINT_TRANSITIONS printf("Previous State: %c Cmd: %.2x State: %c\n", portValuePrevState, command, portValueState); //printf("Event: %c \n", portValueIntHandler);

void StateMachineTest();

// ********************
// * Main entry point *
// ********************
int _tmain(int argc, _TCHAR* argv[])
{
    stateMachine.Initialization = &Initialization;						// state machine initialization
    stateMachine.Initialization( &stateMachine ) ;
	stateMachine.StateInitialization( );								// states initialization
    stateMachine.TransitionMatrixInitialization( &transitionMatrix );	// initialize state machine with transition matrix

	printf("**********************************\n");
	printf("* start test for first injection *\n");
	printf("**********************************\n");
	command             = 0xFF;											// set command to 0xFF (for 0xFF details, please look at transition matrix description)
	portValueState      = 0xFF;
    portValueIntHandler = 0xFF;

	transitionMatrixBuffer       = firstInjectionSystemBehavior;		// transition matrix initialization
    transitionMatrix.startState  = &transitionMatrixBuffer[0]  ;		// S0  pointer
    transitionMatrix.buffer      = &transitionMatrixBuffer[1]  ;		// TM  pointer

	StateMachineTest();

	printf("***********************************\n");
	printf("* start test for second injection *\n");
	printf("***********************************\n");
	command                 = 0xFF;											// set command to 0xFF (for 0xFF details, please look at transition matrix description)
	portValueState          = 0xFF;
	portValuePrevState      = 0xFF;
    portValueIntHandler     = 0xFF;
	portValuePrevIntHandler = 0xFF;

	transitionMatrixBuffer       = secondInjectionSystemBehavior;		// transition matrix initialization
    transitionMatrix.startState  = &transitionMatrixBuffer[0]   ;		// S0  pointer
    transitionMatrix.buffer      = &transitionMatrixBuffer[1]   ;		// TM  pointer

	StateMachineTest();

	getchar(); // main thread block
	
	return 0;
}

void StateMachineTest(){
    
	stateMachine.StartProcessing( );
	PRINT_TRANSITIONS;
	
	// rundom generated number between 0x00 and 0x06
	srand (time(NULL));
	command = rand() % 7;

	for(counter = 0x00; counter < NUMBER_OF_ITERATIONS; counter++, command++){
		
		if(command == 0x07) 
			command = rand() % 7;

		// preserve last state and interrupt values
		portValuePrevState      = portValueState     ;
		portValuePrevIntHandler = portValueIntHandler;

		stateMachine.SetCommand(&command);	
		stateMachine.ProcessCommand();

		if(portValuePrevState != portValueState)
			PRINT_TRANSITIONS;
	}
}

